#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/init.h>

#include <linux/skbuff.h>
#include <linux/netlink.h>
#include <linux/atmdev.h>
#include <linux/atm.h>

/* USB Link/Data LED */
#include <linux/delay.h>
#include <asm/amazon_se/amazon_se_gptu.h>

#include <asm/amazon_se/port.h>
/* copy from port_defs.h */
#define PORT_IOC_CALL(ret,port,pin,func) 	\
	ret=amazon_se_port_reserve_pin(port,pin,PORT_MODULE_ID); \
	if (ret == 0) ret=func(port,pin,PORT_MODULE_ID); \
	if (ret == 0) ret=amazon_se_port_free_pin(port,pin,PORT_MODULE_ID);
#define PORT_MODULE_ID	0xff

#define POLL_FREQ	4

#define USB_LED_OFF	0
#define USB_LED_ON	1
#define USB_LED_FLASH	2

static char usb_led_action = USB_LED_OFF;
static wait_queue_head_t usbledPollingWaitQueue;

/* USB driver reloading */
static struct sock *usb_ksock;
static unsigned int usb_upid;

void usb_set_led_on_action(void)
{
	usb_led_action = USB_LED_ON;
	wake_up_interruptible(&usbledPollingWaitQueue);
}

void usb_set_led_off_action(void)
{
	usb_led_action = USB_LED_OFF;
	wake_up_interruptible(&usbledPollingWaitQueue);
}

void usb_set_led_flash_action(void)
{
	usb_led_action = USB_LED_FLASH;
	wake_up_interruptible(&usbledPollingWaitQueue);
}

static void usb_led_init(void)
{
#ifdef CONFIG_AMAZON_SE_LED_GPIO
	int ret = 0;
	PORT_IOC_CALL(ret,1,11,amazon_se_port_clear_altsel0);
	PORT_IOC_CALL(ret,1,11,amazon_se_port_clear_altsel1);
	PORT_IOC_CALL(ret,1,11,amazon_se_port_set_dir_out);
	PORT_IOC_CALL(ret,1,11,amazon_se_port_set_output);
#endif
}

static void usb_set_led_off(void)
{
#ifdef CONFIG_AMAZON_SE_LED_GPIO
	int ret = 0;
	PORT_IOC_CALL(ret,1,11,amazon_se_port_set_output);
#endif
}

static void usb_set_led_on(void)
{
#ifdef CONFIG_AMAZON_SE_LED_GPIO
	int ret = 0;
	PORT_IOC_CALL(ret,1,11,amazon_se_port_clear_output);
#endif
}

void usb_timer_callback(unsigned long arg)
{
	usb_led_action = USB_LED_ON;
	wake_up_interruptible(&usbledPollingWaitQueue);
}

static int USB_LED_Poll(void)
{
	int		usb_timer;
	static int	timer_is_on = 0;

	usb_timer = set_timer(0, POLL_FREQ * 1000, 1, 0, TIMER_FLAG_CALLBACK_IN_IRQ, (unsigned long)usb_timer_callback, 0);

	while(1)
	{
		interruptible_sleep_on(&usbledPollingWaitQueue);
		switch (usb_led_action)	{
		case USB_LED_OFF:
			usb_set_led_off();
			stop_timer(usb_timer);
			timer_is_on = 0;
			break;
		case USB_LED_ON:
			usb_set_led_on();
			if (timer_is_on) {
				stop_timer(usb_timer);
				timer_is_on = 0;
				//For high speed traffic, LED will stay off for a while
				//add delay to let blink effect significantly , leon
				mdelay(40);
			}	
			break;      
		case USB_LED_FLASH:
			if (!timer_is_on) {
				usb_set_led_off();
				start_timer(usb_timer, 0);
				timer_is_on = 1;
			}
			break;   
		default:
			break;
		} 
	}
}

void usb_module_reset(void)
{
	struct sk_buff *skb;
	int status=0;

	if(!find_task_by_pid(usb_upid)){
		printk(KERN_ERR " %s: find_task_by_pid error!!\n", __FUNCTION__);
		return;
	}

	skb = alloc_skb(64, GFP_ATOMIC);
	skb_put(skb, 64);

	usb_set_led_off();

	status = netlink_unicast(usb_ksock, skb, usb_upid, MSG_DONTWAIT);
	if (status <0){
		printk(KERN_ERR " %s: usb_to_user send fail!!\n", __FUNCTION__);
		kfree_skb(skb);
	}
}

static void usb_krcv (struct sock *sk, int len)
{
	struct sk_buff *skb;
	struct nlmsghdr *nlh = NULL;

	do {
		while((skb = skb_dequeue(&sk->receive_queue)) != NULL)
		{
			nlh = (struct nlmsghdr *)skb->data;
			usb_upid = nlh->nlmsg_pid;
		}
	 } while (usb_ksock && usb_ksock->receive_queue.qlen);
}

static int __init kernel_to_user_init(void)
{
        int nRet;

	/* USB driver reloading */
    	usb_ksock = netlink_kernel_create(NETLINK_USBMSG, usb_krcv);
	if(!usb_ksock){
        	printk(KERN_ERR"open usb kernel socket failed!\n");
        	return -1;
    	}
	netlink_set_nonroot(NETLINK_USBMSG, NL_NONROOT_RECV);

	/* USB LED */
	usb_led_init();
	init_waitqueue_head(&usbledPollingWaitQueue);
        usb_led_action = USB_LED_OFF;
	if ((nRet = kernel_thread(USB_LED_Poll, NULL, CLONE_FS | CLONE_FILES)) < 0)
	{
		printk(KERN_ERR "Open USB LED thread failed!! state: %d\n", nRet);
		return -1;
	}

	return 0;
}

module_init(kernel_to_user_init);
EXPORT_SYMBOL(usb_module_reset);
EXPORT_SYMBOL(usb_set_led_off_action);
EXPORT_SYMBOL(usb_set_led_on_action);
EXPORT_SYMBOL(usb_set_led_flash_action);

